# (C) 2020, Schlumberger. Refer to LICENSE.

# For GPU support also see agnostic.py
import os
import numpy
import datetime
class nokeras_load_model(object):
    def __init__(self):
        pass

    def load_model(data):
        print("keras comomand used, but keras module not installed, a null operations is assumed.")
        return data
try:
    from keras.models import load_model
    class keras_load_model(nokeras_load_model):
        def __init__(self):
            pass
        def load_model(self,data):
            return load_model(data)
except ImportError:
    class keras_load_model(nokeras_load_model):
        pass
    
from sklearn.cluster import KMeans

import distpy.calc.agnostic as agnostic

import scipy.signal
import copy
from scipy import ndimage
from scipy.signal import butter, lfilter, freqz
from scipy.signal import hilbert


# To optimize runtimes - alter this boxsize
# This is used by the reduced memory enabled functions
# see reduced_mem() and available_funcs()
# The reduced memory option is particularly for low memory devices
BOXSIZE=400
GPU_CPU = agnostic.agnostic()

# extra functions for image cleanup (i.e. introduced for plotting, but
# all are 1 or 2D data signal processing so belong in here)
# https://stackoverflow.com/questions/21030391/how-to-normalize-an-array-in-numpy

def normalized(a, axis=-1, order=2):
    xp = GPU_CPU.get_numpy(a)
    l2 = xp.atleast_1d(xp.linalg.norm(a, order, axis))
    l2[l2==0] = 1
    return a / xp.expand_dims(l2, axis)
'''
despike - despiking using a local median filter
'''
def despike(data, m = 2.):
    # Median means that this is not available for GPU
    d = numpy.abs(data - numpy.median(data))
    mdev = numpy.median(d)
    s = d/mdev if mdev else 0.
    return data[s<m]
'''
intensity - rescale data to lie in the 0-1 range
'''
def intensity(data):
    xp = GPU_CPU.get_numpy(data)
    # min goes to zero
    val1 = xp.min(data)
    data = data - xp.min(data)
    # max goes to nearly 1
    data = data / (xp.max(data)+1e-6)
    return data

'''
Curve fitting functions
'''
#various functions for curve fitting
def exp_fit(x, a, b, c):
    xp = GPU_CPU.get_numpy(x)
    return a * xp.exp(b *x) + c
def sum2exp_fit(x, a, b, c, d, e):
    xp = GPU_CPU.get_numpy(x)
    return a*xp.exp(b*x) + c*GPU_CPU.exp(d*x) + e
def lin_fit(x,a,b):
    return a*x + b
# A selection of a curve fit function
def select_fit(strFit, values, initslopes):
    # NOT GPU READY...
    func_fit = lin_fit
    coeffs0 = [initslopes[0],values[-1]]
    if strFit=="exp_fit":
        coeffs0 = [values[0]-values[-1],initslopes[0],values[-1]]
        func_fit = exp_fit
    if strFit=="sum2exp_fit":
        coeffs0 = [values[0]-values[-1],initslopes[0],values[0]-values[-1],initslopes[1],values[-1]]
        func_fit = sum2exp_fit
    return func_fit, coeffs0

'''
 Some functions that are GPU_CPU and used directly in
 pub_command_set.py. This means pub_command_set depends on extra_numpy
 and does not need to know about GPU/CPU issues.
'''
def agnostic_abs(x):
    xp = GPU_CPU.get_numpy(x)
    return xp.abs(x)

def agnostic_angle(x):
    xp = GPU_CPU.get_numpy(x)
    return xp.angle(x)

def agnostic_argmax(x, axis=None):
    xp = GPU_CPU.get_numpy(x)
    return xp.argmax(x, axis=axis)

def agnostic_conj(x):
    xp = GPU_CPU.get_numpy(x)
    return xp.conj(x)

def agnostic_diff(x, n=1, axis=-1):
    xp = GPU_CPU.get_numpy(x)
    return xp.diff(x,n=n,axis=axis, append=0)


def agnostic_unwrap(x, axis=-1):
    xp = GPU_CPU.get_numpy(x)
    return xp.unwrap(x,axis=axis)

def agnostic_roll(x, shift, axis=None):
    xp = GPU_CPU.get_numpy(x)
    return xp.roll(x,shift,axis=axis)



def agnostic_fft(x,axis=-1):
    xp = GPU_CPU.get_numpy(x)
    return xp.fft.fft(x,axis=axis)


def agnostic_ifft(x,axis=-1):
    xp = GPU_CPU.get_numpy(x)
    return xp.fft.ifft(x,axis=axis)

def agnostic_mean(x, axis=None, keepdims=True ):
    xp = GPU_CPU.get_numpy(x)
    return xp.mean(x,axis=axis, keepdims=keepdims)

def agnostic_real(x):
    xp = GPU_CPU.get_numpy(x)
    return xp.real(x)

def agnostic_std(x, axis=None, keepdims= True):
    xp = GPU_CPU.get_numpy(x)
    return xp.std(x,axis=axis, keepdims=keepdims)

def agnostic_sum(x, axis=None, keepdims= True):
    xp = GPU_CPU.get_numpy(x)
    return xp.sum(x,axis=axis, keepdims=keepdims)

def agnostic_zeros(x,size_tuple,dtype=numpy.double):
    xp = GPU_CPU.get_numpy(x)
    return xp.zeros(size_tuple,dtype=dtype)

def agnostic_rescale(x,external_scaling=None):
    xp = GPU_CPU.get_numpy(x)
    minx = xp.amin(x)
    maxx = xp.amax(x)
    if external_scaling is not None:
        minx = xp.amin(external_scaling)
        maxx = xp.amax(external_scaling)
    return (x-minx)/(maxx-minx)

'''
 agnostic_save : allows numpy.save and cupy.save
   Design decision to place this IO function here rather than in io_help.io_helpers

   The older io_helper.numpy_out() is deprecatd and may be removed at a later date
'''
def agnostic_save(stem_name,fname,x):
    xp=GPU_CPU.get_numpy(x)
    xp.save(os.path.join(stem_name,fname+'.npy'),x)

'''
 agnostic_load : allows a generic load - for arbitrary masks and filters.
'''
def agnostic_load(stem_name,fname,x):
    xp=GPU_CPU.get_numpy(x)
    return xp.load(os.path.join(stem_name,fname))

def soft_threshold(x,threshold,direction='greater'):
    if direction=='>':
        x[x>threshold]=threshold
    else:
        x[x<threshold]=threshold
    return x

def hard_threshold(x,threshold, y ,direction='greater'):
    if direction=='>':
        x[x>threshold]=y
    else:
        x[x<threshold]=y
    return x

    
'''
  Extract the analytic signal for a give frequency within a
  narrowband (MLE method)

  It is assumed that the data input is fft'd
'''
def complex_estimator(xp, nx, fc, fs, MLE_length):
    win_len = MLE_length
    n = numpy.arange(0,win_len)
    n = n-(0.5*win_len)
    n = n/fs
    win_hann = xp.hanning( win_len)
    A_line = xp.zeros(nx, dtype=numpy.csingle)
    A_line[0:win_len] = xp.exp( (0+1j) * (2 * numpy.pi * fc * n)) * win_hann
    # This is not zero phase, so we undo the shift that the filter causes...
    A_line = xp.roll(A_line, [-int(round(win_len/2))], axis=0)
    # Estimator (2d ndarray)   
    return A_line

def agnostic_analytic_signal(x,fc,fs,window_length=64,axis=1):
    xp = GPU_CPU.get_numpy(x)
    nx = x.shape[axis]
    est = complex_estimator(xp,nx,fc,fs,window_length)
    est = agnostic_fft(est)
    if axis==1:
        return xp.multiply(x,xp.expand_dims(est, axis=0))
    return xp.multiply(x,est)
    
    
'''
Configure the calculation for different hardware sizes.
'''
def set_boxsize(newval):
    global BOXSIZE
    BOXSIZE = newval

def to_gpu(data):
    return GPU_CPU.asarray(data)

def from_gpu(data):
    return GPU_CPU.asnumpy(data)

'''
Running mean calculation
'''
# see https://stackoverflow.com/questions/13728392/moving-average-or-running-mean
def running_mean_1d(x, N):
    # NOT GPU because of numpy.insert()
    cumsum = numpy.cumsum(numpy.insert(x, 0, 0)) 
    return (cumsum[N:] - cumsum[:-N]) / float(N)

def running_mean(data, N, axis=0):
    # NOT GPU because of extra_numpy.running_mean_1d
    if axis==0:
        for a in range(data.shape[0]):
            data[a,:-N+1]=running_mean_1d(data[a,:],N)
    else:
        for a in range(data.shape[1]):
            data[:-N+1,a]=running_mean_1d(data[:,a],N)
'''
 gather - gather together a list of matrices in a single numpy matrix
'''
def gather(data,prevstack):
    xp = GPU_CPU.get_numpy(data)
    nx = data.shape[0]
    nt = data.shape[1]
    for dataset in prevstack:
        if dataset.ndim > 1:
            nt += dataset.shape[1]
        else:
            nt += 1
    result = xp.zeros((nx,nt),dtype=data.dtype)
    nt = data.shape[1]
    result[:,:nt] = data[:,:]
    ntold=nt
    for dataset in prevstack:
        if dataset.ndim > 1:
            nt += dataset.shape[1]
            result[:,ntold:nt]=dataset[:,:]
        else:
            result[:,ntold:nt]=xp.reshape(dataset,(-1,1))
            nt += 1
        ntold=nt
    return result

'''
 kmeans_clustering : kmeans clustering. Simplest use of KMeans.
'''
def kmeans_clustering(data, n_clusters=10):
    cluster = KMeans(n_clusters=n_clusters).fit(data)
    return cluster.labels_

'''
 interp - a wrapper around numpy's 1d interpolation
'''
def interp(data, xp, fp, axis=0):
    # NOT GPU because of numpy.interp()
    if axis==0:
        for a in range(data.shape[0]):
            data[a,:]=numpy.interp(data[a,:],xp,fp)
    else:
        for a in range(data.shape[1]):
            data[:,a]=numpy.interp(data[:,a],xp,fp)


'''
 approx_vlf - returns a weighted average of the data
              This is a Gaussian window sum, we return the median
              a combination which seems robust to data with fading-lines
'''
def approx_vlf(data):
    # NOT GPU because of scipy.signal
    # A gaussian window to make a weighted average trace, (i.e. a low pass filter)
    win_len = data.shape[1]
    variance = 3000000
    win_gauss = numpy.arange(0,win_len)
    win_gauss = numpy.exp((-(win_gauss-(0.5*win_len))**2)/(2*variance))
    win_gauss = win_gauss/numpy.sum(win_gauss)
    
    # Applied with a median filter (some DAS have fading issues)
    return scipy.signal.medfilt(numpy.average(data,weights=win_gauss,axis=1),3)



'''
 te_from_fft - this is the generic Frequency Band Energy in total-energy form
                 Note that the data is already FFT'd and only the frequencies you
                 want to sum are sent in.
'''
def te_from_fft(localData):
    xp = GPU_CPU.get_numpy(localData)
    # calulate Total Energy (sum of squares)
    nx = localData.shape[0]
    squared = xp.real(localData*xp.conj(localData))
    nval = (len(squared[0])-1)*2
    # integration for RMS
    if nval%2:
        rmsval = (squared[:,0]+2*xp.sum(squared[:,1:],axis=1))
    else:
        rmsval = (squared[:,0]+2*xp.sum(squared[:,1:-1],axis=1) + squared[:,-1])
    return rmsval

'''
 rms_from_fft - this is the generic Frequency Band Energy in root-mean-square form
                 Note that the data is already FFT'd and only the frequencies you
                 want to sum are sent in.
'''
# The general FBE calculation - the sender must control which part of the FFT is sent in...
def rms_from_fft(localData):
    xp = GPU_CPU.get_numpy(localData)
    # calulate RMS
    nx = localData.shape[0]
    nval = (nx-1)*2
    # First get the total energy
    rmsval = te_from_fft(localData)
    # Convert to RMS for output
    return (xp.sqrt(rmsval/nval)/xp.sqrt(nval))


'''
 macro_factory - a mini command factory.
     This allows for the definition of macros. So more complex commands are
     encapsulated in a single command - see pub_command_set.Macro
'''
def macro_factory(args, traces):
    # NOTE GPU because of scipy.ndimage.median_filter
    # set up all the possible defaults...
    axis_val = args.get('axis',0)
    distance_val = args.get('distance',5)
    sigma_val = args.get('sigma',2)
    zone_index = args.get('zone_index',0)
    threshold = args.get('threshold',0)
    suffix = args.get('suffix','POST')

    commandList = {
        "abs"             : numpy.abs(traces),
        "gaussian_filter" : ndimage.gaussian_filter(traces,sigma=sigma_val),
        "gradient"        : numpy.gradient(traces.astype(numpy.double), distance_val, axis=axis_val),
        "normalize"       : normalized(traces, axis=axis_val),
        "percentage_on"   : 100*numpy.sum(traces,axis=axis_val)/traces.shape[axis_val],  
        "sum"             : numpy.sum(traces, axis=axis_val),
        "threshold"       : (traces > threshold),
        "median_filter"   : ndimage.median_filter(traces, distance_val)
    }
    # if the command is not found the data is returned untouched...
    return commandList.get(args['name'], traces)
    
'''
 reduced_mem - manages memory for processing on low spec Edge devices
             - Usage: reduced_mem(dataIn, dataout, te_from_fft, {})
             -        reduced_mem(dataIn, dataOut, numpy.fft.fft, {'axis' : 1}
'''
def reduced_mem(dataIn, dataOut, func, keyword_arguments):
    global BOXSIZE
    nx = dataIn.shape[0]
    istart = 0    
    while istart+BOXSIZE<nx:
        dataOut[istart:istart+BOXSIZE,:] = func(dataIn[istart:istart+BOXSIZE,:], **keyword_arguments).reshape(BOXSIZE,dataOut.shape[1])
        istart = istart + BOXSIZE
    if istart<nx:
        dataOut[istart:,:] = func(dataIn[istart:,:], **keyword_arguments).reshape(dataOut.shape[0]-istart,dataOut.shape[1])
        istart = istart + BOXSIZE
    return dataOut

'''
 up_wave - performs a 2D FFT, separates the up-going wave and returns the 2D iFFT
         - NOTE: the caller will end up with a complex matrix of numbers.
'''
def up_wave(data):
    # Assume 2D FFT
    SMALL=1.0e-6
    m = len(data)
    n = len(data[0])
    m_on_2 = m//2
    n_on_2 = n//2
    DATA = copy.deepcopy(data)

    DATA[m_on_2:,:n_on_2]*=SMALL
    DATA[:m_on_2,n_on_2:]*=SMALL
    DATA[:m_on_2,n_on_2:]*=(1-SMALL)
    DATA[m_on_2:,:n_on_2]*=(1-SMALL)
    xp = GPU_CPU.get_numpy(DATA)
    return (xp.fft.ifft2(DATA))

'''
 down_wave - performs a 2D FFT, separates the up-going wave and returns the 2D iFFT
           - NOTE: the caller will end up with a complex matrix of numbers.
           see also - up_wave
'''
def down_wave(data):
    # Assume 2D FFT 
    SMALL=1.0e-6
    m = len(data)
    n = len(data[0])
    m_on_2 = m//2
    n_on_2 = n//2

    DATA = copy.deepcopy(data)
    DATA[m_on_2:,:n_on_2]*=(1-SMALL)
    DATA[:m_on_2,n_on_2:]*=(1-SMALL)
    DATA[:m_on_2,n_on_2:]*=SMALL
    DATA[m_on_2:,:n_on_2]*=SMALL
    xp = GPU_CPU.get_numpy(DATA)
    return (xp.fft.ifft2(DATA))

'''
 source - Usage: pass in the results of up_wave and down_wave as up and down
        -        The returned matrix highlights sources
        -       (i.e. the cross-corrlation of the up and down wave is maximum at sources)
'''
def source(up,down):
    xp = GPU_CPU.get_numpy(up)
    return xp.abs((up*xp.conj(down)))

'''
 reflection - Usage: pass in the results of up_wave and down_wave as up and down
            -        The returned matrix highlights reflections
            -        (i.e. the down-data deconvolved from the up-data
'''
def reflection(up,down):
    xp = GPU_CPU.get_numpy(up)
    TOLERANCE = 1.0e-7
    return xp.abs((up*xp.conj(down))/((down*xp.conj(down))+TOLERANCE))

'''
 vel_map - Returns the phase-velocity map in the 2D FFT space
'''
def vel_map(nx,nt,dx,dt):
    # COULD BE GPU COMPATIBLE...BUT VEL_MASK IS NOT
    TOLERANCE=1.0e-7
    freq = numpy.linspace(0,1.0/(2*dt),nt//2)
    wavn = numpy.linspace(0,1.0/(2*dx),nx//2)
    frequency = numpy.zeros((1,nt),dtype=numpy.double)
    frequency[0,:nt//2]=freq
    frequency[0,-(freq.shape[0]):]=numpy.flipud(freq)
    halfLength = nx//2
    wavenumber = numpy.zeros((nx,1),dtype=numpy.double)
    wavenumber[0:halfLength]=numpy.reshape(wavn,(halfLength,1))
    lineTest = numpy.reshape(numpy.flipud(wavn),(halfLength,1))
    wavenumber[halfLength:]= lineTest[:len(wavenumber[halfLength:])]
    freqImage = numpy.reshape(numpy.tile(frequency,(nx,1)),(nx,nt))
    wavnImage = numpy.reshape(numpy.tile(wavenumber,(1,nt)),(nx,nt))
    return freqImage/(wavnImage+TOLERANCE)

'''
 vel_mask - use the results of vel_map to create a filter that band passes phase velocities
        - between vel_low and vel_high. The smoothing is applied so that the edges are not too
        - sharp. The padtype='even' prevents strong filter edge effects.
'''
def vel_mask(velmap,vel_low,vel_high,smooth=0):
    # NOT GPU COMPATIBLE DUE TO scipy.signal.filtfilt
    nx=velmap.shape[0]
    nt=velmap.shape[1]
    velm = numpy.zeros((nx,nt),dtype=numpy.double)
    velm[numpy.where(velmap > vel_low)]=1.0
    velm[numpy.where(velmap > vel_high)]=0.0
    if smooth>1:
        Wn = 1.0/smooth
        b, a = scipy.signal.butter(5,Wn,'low',output='ba')
        velm = numpy.abs(scipy.signal.filtfilt(b,a,velm,axis=1, padtype='even'))
    return velm
    
'''
  multiple_calcs - allows a range of FBEs to be calculated using the reduced_mem calculation system
'''
def multiple_calcs(localData, low_freq, high_freq, func):
    xp = GPU_CPU.get_numpy(localData)    
    bulkOut = xp.zeros((localData.shape[0],len(low_freq)),dtype=numpy.double)
    oneOut = xp.zeros((localData.shape[0],1))
    for a in range(len(low_freq)):
        oneOut = reduced_mem(localData[:,low_freq[a]:high_freq[a]],oneOut, func, {})
        bulkOut[:,a]=numpy.squeeze(oneOut)
    return bulkOut

'''
 running_max - return a running maximum value
'''
def running_max(data, running_window_length):
    nx = len(data)
    nt = len(data[0])
    xp = GPU_CPU.get_numpy(data)
    result = xp.zeros((nx,nt),dtype=xp.double)
    for a in range(0,nx,running_window_length):
        result[a,:]=xp.max(data[a:a+running_window_length],axis=0)
    return result

'''
 running_min : return a runningin minimum value
'''
def running_min(data, running_window_length):
    nx = len(data)
    nt = len(data[0])
    xp = GPU_CPU.get_numpy(data)
    result = xp.zeros((nx,nt),dtype=xp.double)
    for a in range(0,nx,running_window_length):
        result[a,:]=xp.min(data[a:a+running_window_length],axis=0)
    return result

'''
 sta_lta : short-term-average over long-term-average, a good signal onset detection transform
'''
def sta_lta(data, sta, lta):
    # Forward looking sta, backward looking lta...
    nx = len(data)
    nt = len(data[0])
    # NOT GPU because of ndimage.uniform_filter
    sta_result = ndimage.uniform_filter(data,size=(1,sta),mode='nearest')/sta
    lta_result = ndimage.uniform_filter(data,size=(1,lta),mode='nearest')/lta
    lta_result[numpy.where(numpy.abs(lta_result)<1e-6)]=1e-6
    return (sta_result/lta_result)

'''
 peak_to_peak : find the largest range occurring within the provided window
'''
def peak_to_peak(data,window_length):
    # NOT GPU because of ndimage.maximum_filter
    nx = len(data)
    nt = len(data[0])
    half_length=int(window_length*0.5)
    return numpy.max(ndimage.maximum_filter(data,size=(1,window_length),mode='nearest')-ndimage.minimum_filter(data,size=(1,window_length),mode='nearest'),axis=1)

'''
 count_peaks : count the number of peaks using scipy.signal.find_peaks_cwt
'''
def count_peaks(data,sta,lta):
    # NOT GPU because of scipy.signal.find_peaks_cwt
    nx = len(data)
    nt = len(data[0])
    result=numpy.zeros((nx,1),dtype=numpy.int)
    for a in range(nx):
        peak_kind = scipy.signal.find_peaks_cwt(data[a,:],numpy.arange(sta,lta,5))
        result[a]=len(peak_kind)
    return result

'''
 rescaled_freq : used to turn a frequency into the nearest index in the Fourier space
'''
def rescaled_freq(freq_list, freq_rescale):
    newlist=[]
    for freq in freq_list:
        newlist.append(int(freq*freq_rescale))
    return newlist


'''
virtual_source - assumes that the data is in the frequency domain

The returned result is still in the frequency domain
'''
def virtual_source(data,traceId):
    nx = data.shape[0]
    nt = data.shape[1]
    xp = GPU_CPU.get_numpy(data)
    sourceImage = xp.reshape(xp.tile(data[traceId,:],(nx,1)),(nx,nt))
    return data*xp.conj(sourceImage)

'''
the seismic sweetness attribute is the amplitude of the analytic signal
divided by the sqrt of the phase. It highlights high amplitude at low
frequency (in young clastic sedimentary basins this gives "sweet spots"
for potential hydrocarbons).
'''
def sweetness(data, dx):
    dk = 1/dx
    # NOT GPU because of scipy.signal.hilbert
    analytic_signal = hilbert(data,axis=0)
    instantaneous_phase = numpy.unwrap(numpy.angle(analytic_signal))
    instantaneous_frequency = (numpy.diff(instantaneous_phase) /(2.0*numpy.pi*dk)) 
    return numpy.abs(analytic_signal)/numpy.sqrt(instantaneous_frequency+1e-7)

'''
The scattering matrix - given Up and Down.
 | Rdown  Tup |
 | Tdown  Rup |

expects the fft(up) and fft(down) as inputs as
this is a frequency domain evaluation
'''
def scattering_matrix(up, down, ispan):
    Pup = up
    Pdown = down
    xp = GPU_CPU.get_numpy(up)
    Qup = xp.roll(up,ispan,axis=0)
    Qdown = xp.roll(down,ispan,axis=0)
    denominator = (Pdown*xp.conj(Qdown))-(xp.conj(Pup)*Qup)
    # threshold of 100 - seem high???
    denominator = xp.sign(denominator)/(xp.abs(denominator)+100.0)

    # scattering matrix
    Rdown=(Pup*xp.conj(Qdown))  -(xp.conj(Pdown)*Qup)
    Tup  =(Pdown*xp.conj(Pdown))-(Pup*xp.conj(Pup))
    Tdown=(Qdown*xp.conj(Qdown))-(Qup*xp.conj(Qup))
    Rup  =(Pdown*xp.conj(Qup))  -(xp.conj(Pup)*Qdown)
    # denominator correction
    Rdown = Rdown*denominator
    Tup   = Tup*denominator
    Tdown = Tdown*denominator
    Rup   = Rup*denominator

    return (Rdown, Tup, Tdown, Rup)

'''
 destripe : does a good job of removing "phase skips" that are vertical lines.
            See Bouali, (2010). A simple and robust destriping algorithm for
            imaging spectrometers: application to modis data. San Diego, California,
            ASPRS 2010 Annual Conference.
'''
def destripe(localData):
    # Vertical destriping...
    TOLERANCE=1.0e-6
    dt=1
    dx=1
    xp = GPU_CPU.get_numpy(localData)
    nx = len(localData)
    nt = len(localData[0])
    freq = xp.linspace(0,1.0/(2*dt),xp.int(nt/2))
    frequency = xp.zeros((1,nt),dtype=xp.double)
    frequency[0,0:xp.int(nt/2)]=freq
    frequency[0,-xp.int(nt/2):]=xp.flipud(freq)
    wavn = xp.linspace(0,1.0/(2*dx),xp.int(nx/2))
    wavenumber = xp.zeros((nx,1),dtype=xp.double)
    wavenumber[0:xp.int(nx/2)]=xp.reshape(wavn,(xp.int(nx/2),1))
    wavenumber[-xp.int(nx/2):]=numpy.reshape(xp.flipud(wavn),(xp.int(nx/2),1))
    freqImage = xp.reshape(xp.tile(frequency,(nx,1)),(nx,nt))
    wavnImage = xp.reshape(xp.tile(wavenumber,(1,nt)),(nx,nt))
    #Rx = (freqImage*freqImage)/((freqImage*freqImage)+(wavnImage*wavnImage)+TOLERANCE)
    #Rx[Rx<1e-6]=1e-6
    Ry = (wavnImage*wavnImage)/((freqImage*freqImage)+(wavnImage*wavnImage)+TOLERANCE)
    Ry[Ry<1e-6]=1e-6
    return xp.real(xp.fft.ifft2(xp.fft.fft2(localData)*Ry))

'''
 acoustic_properties : the eigenvalues of the scattering matrix
'''
def acoustic_properties(Rdown,Tup,Tdown,Rup):
    xp = GPU_CPU.get_numpy(Tup)
    tl=1-(xp.conj(Rdown)*Rdown+Tdown*xp.conj(Tdown))
    tr=(xp.conj(Rdown)*Tup)+(xp.conj(Tdown)*Rup)
    bl=(xp.conj(Tup)*Rdown)+(xp.conj(Rup)*Tdown)
    br=1-((xp.conj(Tup)*Tup)+(xp.conj(Rup)*Rup))
    detS=(tl*br)-(tr*bl)
    traceS=tl+br
    eval1=xp.real(0.5*(traceS+sqrt((traceS**2)-(4*detS))))
    eval2=xp.real(0.5*(traceS-sqrt((traceS**2)-(4*detS))))
    return (eval1, eval2)

'''
log_abs_sq : useful for computing the cepstrum if data is already in the frequency domain
             cepstrum is
                    abs(ifft(log_abs_sq(data)))
             so it can be calculated by a worklflow:
                    fft -> log_abs_sq -> ifft -> abs
'''
def log_abs_sq(data):
    xp = GPU_CPU.get_numpy(data)
    return xp.log(xp.abs(data)**2)

'''
 hash : take the next 64 values above and below 0.5 (assume the previous calculation
        has left the values as 0 and 1.
        Convert to a single 64-bit integer
'''
def hash(data,ioffset=0):
    xp = GPU_CPU.get_numpy(data)
    outData = xp.zeros((data.shape[0],1),dtype=xp.uint64)
    for a in range(data.shape[0]):
        outData[a] = sum(int(j)<<int(i) for i,j, in enumerate(reversed((data[a,ioffset:ioffset+64]).flatten().tolist())))
    return outData

'''
  keras_model : load a saved keras model for evaluation or further training
'''
def keras_model(data,kerasfile,Y=None,train=None):
    # NULL behaviour is to simply return the input
    if kerasfile is None:
        return data
    kclass = keras_load_model()
    model = kclass.load_model(kerasfile)
    if model is not None:
        if train is not None:
            epochs = train.get('epochs',150)
            batch_size = train.get('batch_size',10)
            model.fit(data,Y,epochs=epochs,batch_size=batch_size,verbose=0)
            model.save(kerasfile)
            model.evaluate(data,Y,verbose=0)
        return model.predict(data)
    return data
'''
 available_funcs : these are the functions for which the reduced_mem options are available.
'''
def available_funcs(funcName):
    if funcName=='te_from_fft':
        return te_from_fft
    if funcName=='rms_from_fft':
        return rms_from_fft
    if funcName=='fft':
        return agnostic_fft




# (c) 2020, Schlumberger. Refer to LICENSE

import os
import numpy

dirintest = "{PATH}\\results\\vlf"

'''
  dirswap : a utility for converting stage 1 (strainrate-to-summary)
            results (1 file per second) to
            inputs (all seconds in 1 file) for stage 2 (summary-to-plots)
            
  Swaps the structure so that the numpy files are named by their directory
  and so that the filenames become the directory names.


 e.g. the structure
 1535463913\band_0-nyquist.npy
 1535463913\band_01_0-1.npy
 1535463913\measured_depth.npy
 1535463913\time.npy
 1535463923\band_0-nyquist.npy
 1535463923\band_01_0-1.npy
 1535463923\measured_depth.npy
 1535463923\time.npy
 becomes...
 band_0-nyquist\1535463913.npy
 band_0-nyquist\1535463923.npy
 band_01_0-1\1535463913.npy
 band_01_0-1\1535463923.npy
 measured_depth\1535463913.npy
 measured_depth\1535463923.npy
 time\1535463913.npy
 time\1535463923.npy

 This then means that you can operate on the *.npy files, for example
 for generating plots or analyzing perceptual hashes.
'''
def dirswap(dirin):
    MD_NAME = 'measured_depth.npy'
    dirnames = [d for d in os.listdir(dirin) if os.path.isdir(os.path.join(dirin,d))]
    testdir = os.path.join(dirin,dirnames[0])
    fnames = [f for f in os.listdir(testdir) if os.path.isfile(os.path.join(testdir,f))]
    # make directories for the results
    xaxis=None
    for filename in fnames:
        if filename[-3:]=='npy':
            if filename==MD_NAME:
                xaxis=numpy.load(os.path.join(testdir,filename))
            dirname = filename[:-4]
            #fulldir = os.path.join(dirin,dirname)
            #if not os.path.exists(fulldir):
            #    os.makedirs(fulldir)
    for dirname in dirnames:
        testdir = os.path.join(dirin,dirname)
        fnames = [f for f in os.listdir(testdir) if os.path.isfile(os.path.join(testdir,f))]
        for filename in fnames:
            if filename[-3:]=='npy':
                dirout = os.path.join(dirin,filename[:-4])
                data = numpy.load(os.path.join(testdir,filename))
                #numpy.save(os.path.join(dirout,dirname+'.npy'), data)
                numpy.save(os.path.join(dirout+'.npy'), data)                
                #if not os.path.exists(os.path.join(dirout,MD_NAME)):
                #    numpy.save(os.path.join(dirout,MD_NAME),xaxis)
                # remove the old file...
                print('removing ',os.path.join(testdir,filename))
                os.remove(os.path.join(testdir,filename))

# testing
if __name__ == "__main__":
    dirswap(dirintest)



# (c) 2020 Schlumberger. Refer to LICENSE

# directory services - in the short term this is just a wrapper on
#   some standard os functions.
#   It supports a storage_type identifier, as some cloud storage types
#   do not support python's os (POSIX) naming. Some examples for Google
#   Cloud Buckets using blobs rather than *.npy files are coded for reference.
#
import os
import numpy
import string
import csv

BACKSCATTER_BUCKET = "google-storage-cloud-address"

'''
 path_join : a wrapper for os.path.join()
             The storage_type default uses os.path.join
'''
def path_join(dirname1, dirname2, storage_type=1):
    if storage_type==1:
        return os.path.join(dirname1,dirname2)
    else:
        return os.path.join(dirname1,dirname2)
'''
 path_join : a wrapper for os.path.join()
             The storage_type default uses os.path.join
'''
def exists(pathname):
    if not os.path.exists(pathname):
        return False
    return True

'''
 makedir : a wrapper for os.path.makedirs generating the path if it does not
           exist.

           Current question: For Google Blob storage the naming convention can
           look likea a directory\file so does this just pass?
'''
def makedir(dirname):
    if not os.path.exists(dirname):
        os.makedirs(dirname)

'''
 load : wrappers numpy.load() for filestores, otherwise interprets a byte-based
        blob from a google cloud bucket.
'''
def load(filename, storage_type=1):
    if storage_type == 1:
        return numpy.load(filename)
    else:
        #rawObj = blob_server(filename,PROVISIONED)
        bucket = client.get_bucket(BACKSCATTER_BUCKET)
        blob = bucket.get_blob(filename)
        byteval = blob.download_as_string()
        time_ref = numpy.frombuffer(byteval, dtype=numpy.double)
        return time_ref

# read zones - used to read cluster and perforation locations
# from csv files. Notethis is NOT used to read DTS data from CSV format files!
def csv_reader(filename, storage_type=1):
    lines = []
    reader = None
    if storage_type==1:
        with open(filename) as csvFile:
            reader = csv.reader(csvFile, delimiter=' ')
            for row in reader:
                lines.append(row)
    else:
        bucket = client.get_bucket(BACKSCATTER_BUCKET)
        blob = bucket.get_blob(filename)
        reader = csv.reader(io.StringIO((blob.download_as_string()).decode("utf-8")), delimiter=' ')
        for row in reader:
            lines.append(row)
    return lines






# (C) 2020, Schlumberger. Refer to LICENSE


########################################
# USER CONFIGURABLE VALUES:
# h5 formats - note that PRODML h5 is not universally adopted
#
# Note that these functions are based on reading an HDF5 file, and not defined from any standard
# It is fairly generic and can be adapted to the different HDF5 formats of other
# vendors.

filenameH5 = "{PATH}\\h5\\das_data.h5"
basepath = "{PATH}\\data"
WRITE_DATA = False
########################################

import h5py
import os
import numpy
import distpy.io_help.io_helpers
import distpy.io_help.directory_services as ds


'''
 list_keys : recursively list the keys from an h5 file
'''
def list_keys(root, basepath):
    for value in root.values():
        try:
            print(value.name,'   ',value.shape)
        except:
            print(value.name)
            list_keys(value, basepath)
        finally:
            pass

'''
 read_h5 : recursively try to dump matrices and arrays as numpy *.npy files - blindly...
'''
def read_h5(root, basepath):
    for value in root.values():
        try:
            # failure is based on value.shape throwing an exception in the print statement
            # ...so do not remove this print statement
            print(value.name,'   ',value.shape)
            numpy.save(os.path.join(basepath,value.name),value)
        except:
            try:
                dirname = basepath+value.name
                print(value.name)
                ds.makedir(dirname)
                ingest_h5(value, basepath)
            except:
                print('FAILED at ',value.name)
                pass
            finally:
                pass
        finally:
            pass

'''
 type0_h5  : a first example of hdf5 import.
             Reads an h5 file. First version supports files that write 1-second
             chunks in a 'data','depth,'time' format.
'''
def type0_h5(readObj, basepath, filename):
    # filename contains datestamp...
    # 161215_053839.h5
    # 1234567890123456
    endFname = filename[-16:]
    day = int(endFname[:2])
    month = int(endFname[2:4])
    year = int(endFname[4:6])
    hour = int(endFname[7:9])
    minute = int(endFname[9:11])
    second = int(endFname[11:13])
    timeObj = dtime(2000+year,month,day,hour,minute,second)
    unixtime = timeObj.timestamp()
    print(str(unixtime))


    numpy.save(os.path.join(basepath,'measured_depth.npy'),readObj['depth'][()])
    total_traces = readObj['data'].shape[0]
    chunksize = readObj['data'].chunks[0]
    increment=0
    for a in range(0,total_traces,chunksize):
        fname = str(int(unixtime+increment))+'.npy'
        fullPath = os.path.join(basepath,fname)
        if not os.path.exists(fullPath):
            numpy.save(fullPath,readObj['data'][a:a+chunksize,:].transpose())
        else:
            print(fullPath,' exists, assuming restart run and skipping.')
        increment+=1

'''
 type1_h5  : A common HDF5 format used in DAS
     depth = '/Acquisition/FacilityCalibration[0]/Calibration[0]/LocusDepthPoint' in readObj
     time  = '/Acquisition/Raw[0]/RawDataTime' in readObj
     data  = '/Acquisition/Raw[0]/RawData' in readObj

     This example is given in https://github.com/Schlumberger/distpy/wiki/Dev-Tutorial-:-Extending-distpy-with-new-ingesters
'''
def type1_h5(readObj, basepath):
    depths = readObj['/Acquisition/FacilityCalibration[0]/Calibration[0]/LocusDepthPoint']
    times  = readObj['/Acquisition/Raw[0]/RawDataTime']
    
    # The data can be large - so don't read it all into memory at once.
    data_name = '/Acquisition/Raw[0]/RawData'

    time_vals = numpy.zeros((times.shape[0]))
    for a in range(times.shape[0]):
        time_vals[a] = times[a]*1e-6
    prf = int(numpy.round(1.0/numpy.mean(time_vals[1:]-time_vals[:-1])))


    depth_vals = numpy.zeros((depths.shape[0]))
    for a in range(depths.shape[0]):
        tuple_val = depths[a]
        depth_vals[a] = tuple_val[1]
    numpy.save(os.path.join(basepath,'measured_depth.npy'),depth_vals)

    unixtime = int(time_vals[0])
    print(str(unixtime))

    # How many 1 second files to write from this HDF5
    nsecs = int(times.shape[0]/prf)
    ii=0
    for a in range(nsecs):
        fname = str(int(unixtime+a))+'.npy'
        fullPath = os.path.join(basepath,fname)
        if not os.path.exists(fullPath):
            numpy.save(fullPath,readObj[data_name][ii:ii+prf,:].transpose())
        else:
            print(fullPath,' exists, assuming restart run and skipping.')
        ii+=1

'''
 ingest_h5 : select the reader that recognizes the file type, and read the file
'''
def ingest_h5(filename,basepath):
    readObj = h5py.File(filename,'r')
    
    # TYPE 0:
    depth = 'depth' in readObj
    #time  = '/Acquisition/Raw[0]/RawDataTime' in readObj
    data  = 'data' in readObj
    if depth and data:
        type0_h5(readObj,basepath,filename)
    
    # TYPE 1:
    depth = '/Acquisition/FacilityCalibration[0]/Calibration[0]/LocusDepthPoint' in readObj
    time  = '/Acquisition/Raw[0]/RawDataTime' in readObj
    data  = '/Acquisition/Raw[0]/RawData' in readObj
    if depth and time and data:
        type1_h5(readObj,basepath)
    readObj.close()


'''
 main : two modes, WRITE_DATA==True - recursively try to dump data
                   WRITE_DATA==False - recursively list the key-value contents of the file

        In general we use WRITE_DATA==False to determine format information, and then construct
        a custom reader. In some cases WRITE_DATA==True is sufficiently close to what we need that
        the custom reader just performs directory renaming. However, note that in blob storage the
        directory rename is a deep copy and usually very slow - so on cloud custom h5 readers that
        set up the output name at the start are really important from a cost perspective.
'''
def main():
    filename = filenameH5
    readObj = h5py.File(filename,'r')
    if WRITE_DATA==False:
        list_keys(readObj,basepath)
    else:
        read_h5(readObj, basepath)
    readObj.close()


if __name__ == "__main__":
    main()


        



# (C) 2020, Schlumberger. Refer to LICENSE

########################################
# USER CONFIGURABLE VALUES:
# A system configuration to optimize performance
jsonSystemConfig = "{PATH}\\config\\systemConfig.json"
# A signal processing workflow configuration
jsonConfig = "{PATH}\\config\\strainrate2summary.json"
########################################
import argparse

import copy
import os
import numpy
from distpy.workers.strainrate2summary import strainrate2summary
import multiprocessing
import distpy.io_help.io_helpers as io_helpers



def main(configOuterFile, extended_list=[]):
    basedir,dirout,jsonConfig,PARALLEL,NCPU,BOX_SIZE, xaxisfile,taxisfile,prf = io_helpers.systemConfig(configOuterFile)

    xaxis = None
    try:
        xaxis = numpy.load(xaxisfile)
    except FileNotFoundError:
        print('No x-axis specified')

    taxis = None
    try:
        taxis = numpy.load(taxisfile)
    except FileNotFoundError:
        taxis = None
    

    configFile = io_helpers.json_io(jsonConfig,1)
    configData = configFile.read()
    # Note this copy - we don't (from a user point-of-view)
    # want to be doing system configuration in the strainrate2summary.json,
    # but we do from an internal point-of-view
    configData['BOX_SIZE']=BOX_SIZE
    configData['taxis'] = taxis
    verbose = configData.get('verbose',0)
    if verbose==1:
        print(configData)
 
    #scan for directories
    datafiles=[]
    for root, dirs, files in os.walk(basedir):
        for datafile in files:
            if not datafile=='measured_depth.npy':
                if not datafile=='time.npy':
                    datafiles.append(os.path.join(root,datafile))
        # break here because we don't want subdirectories (SLB case)
        break

    if not PARALLEL:
        # parallel does not work in Techlog...
        for datafile in datafiles:
            print(datafile)
            strainrate2summary(datafile, xaxis, prf, dirout, configData, copy.deepcopy(extended_list))
    else:
        manager = multiprocessing.Manager()
        q = manager.Queue()
        print(multiprocessing.cpu_count())
        
        pool = multiprocessing.Pool(processes=NCPU, maxtasksperchild=1)


        jobs = []
        for datafile in datafiles:
            job = pool.apply_async(strainrate2summary, [datafile, xaxis, prf,dirout, configData,copy.deepcopy(extended_list)])
            print(job)
            jobs.append(job)

        for job in jobs:
            job.get()
            print(job)

        q.put('kill')
        pool.close()

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-f","--filename", help='A configuration file')
    args = parser.parse_args()
    main(args.filename)



# (C) 2020, Schlumberger. Refer to LICENSE

########################################
# USER CONFIGURABLE VALUES:
inpath = "{PATH}\\las2"
outpath = "{PATH}\\witsml"
PARALLEL=True
NCPU=1
########################################
import argparse

import os
import distpy.io_help.witsmlfbe as witsmlfbe
import multiprocessing

'''
convert : separates the call to the conversion from the main() function
          This allows the main function itself to not depend on distpy.io_help
          the dependency is fielded here
'''
def convert(infile,outfile):
    root=witsmlfbe.las2fbe(infile)
    witsmlfbe.writeFBE(outpath,outfile,root)


####################################################################################################
# Executable code...

def main(configOuterFile):
    # standard configuration information
    inpath,outpath,configFile,PARALLEL,NCPU,BOX_SIZE, xaxisfile,taxisfile,prf = io_helpers.systemConfig(configOuterFile)

    if not os.path.exists(outpath):
        os.makedirs(outpath)

    #################################
    # Process all the files with the
    # .las extension
    #################################
    inlist = []
    outlist = []
    for filename in os.listdir(inpath):
        if os.path.isfile(os.path.join(inpath,filename)):
            if filename[-4:]=='.las':
                inlist.append(os.path.join(inpath,filename))
                outlist.append(filename)

    if not PARALLEL:
        # parallel does not work in Techlog...
        for (infile,outfile) in zip(inlist,outlist):
            convert(infile,outfile)
    else:
        manager = multiprocessing.Manager()
        q = manager.Queue()
        print(multiprocessing.cpu_count())
        
        pool = multiprocessing.Pool(processes=NCPU, maxtasksperchild=1)
        
        jobs = []
        for (infile,outfile) in zip(inlist,outlist):
            job = pool.apply_async(convert, [infile,outfile])
            print(job)
            jobs.append(job)

        for job in jobs:
            job.get()
            print(job)

        q.put('kill')
        pool.close()

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-f","--filename", help='A configuration file')
    args = parser.parse_args()
    main(args.filename)



'''
pub_command_set.py

This module contains the list of public commands. These will generally wrapper some
numpy or extra_numpy function.

(C) 2020, Schlumberger. Refer to LICENSE
'''
import numpy
import os
from scipy import signal
from scipy import ndimage
from scipy import stats
from scipy.optimize import curve_fit
from scipy.signal import butter, lfilter, freqz
from scipy.stats import kurtosis, hmean, gmean, skew, moment
import distpy.calc.extra_numpy as extra_numpy
import distpy.io_help.io_helpers as io_helpers
import distpy.io_help.directory_services as directory_services


'''
  The universal_arglist is used in automatic documentation generation.
  By holding a universal_arglist we are gently encouraging re-use where
  possible and gently avoiding the same argument keyword having very
  different meanings in different contexts.
'''
NONE = "NONE"
DEFAULT = "default"
DESC = "description"
universal_arglist = {
    "args" :          {DEFAULT : None, DESC : "Arguments a passed through to the underlying math library"},
    "coeffs" :        {DEFAULT : None, DESC : "Filter coefficients"},
    "xaxis" :         {DEFAULT : None, DESC : "A numpy vector of distances along the fibre"},
    "taxis" :         {DEFAULT : None, DESC : "A numpy vector of times associated with columns of data"},
    "directory_in" :  {DEFAULT : NONE, DESC : "The subdirectory from which to read data"},
    "directory_out" : {DEFAULT : NONE, DESC : "The subdirectory where results will be written"},
    "direction"     : {DEFAULT : ">", DESC : "The direction for applying the threshold, > or <"},
    "double_ended"  : {DEFAULT : -1,   DESC : "For handling double-ended fibre [-1=single-ended, 0=start-of-fibre half, 1=end-of-fibre half"},
    "edge_order"    : {DEFAULT :  1,   DESC : "Gradient is calculatd using N-th order accurate differences at the boundaries"},
    "filename"      : {DEFAULT : NONE, DESC : "A filename, the complete path will be defined by directory_out and project configurations"},
    "command_list" :  {                DESC : "A list of sub-commands collected in a single macro"},
    "train" :         {                DESC : "A dictionary containing training parameters for keras, e.g. { 'epochs' : 150, 'batch_size' : 10 }"},
    "func" :          {                DESC : "Either rms_from_fft or te_from_fft"}, 
    "labels" :        {                DESC : "A list of column headers"},
    "order" :         {DEFAULT : 5,    DESC : "The order for a filter calculation such as the Butterworth filter"},
    "type" :          {DEFAULT : "lowpass",DESC : "The type of a filter which can be lowpass, highpass, bandpass, or bandstop"},
    "padtype" :       {DEFAULT : "even",DESC : "The type of end-effect control on a filter, see scipy.signal.filtfilt"},
    "prf" :           {DEFAULT : 10000, DESC : "The pulse repetition frequency in Hz (one over the time sample rate)"},
    "freq" :          {DEFAULT : 200.0, DESC : "A frequency in Hz or a wavenumber in 1/m"},
    "filename" :      {DEFAULT : NONE,  DESC : "A filename for read or write operations"},
    "axis" :          {DEFAULT : -1,    DESC : "The axis to apply an operation to, typically in distpy axis=0 for depth and axis=1 for time"},
    "xsigma" :        {DEFAULT : 1.0,   DESC : "A standard deviation in the x-direction"},
    "tsigma" :        {DEFAULT : 1.0,   DESC : "A standard deviation in the time direction"},
    "xorder" :        {DEFAULT : 5,     DESC : "In a 2D filter this is the order in the x-direction"},
    "torder" :        {DEFAULT : 5,     DESC : "In a 2D filter, this is the order in the t-direction"},
    "distance" :      {DEFAULT : 3,     DESC : "The number of samples in a median filter"},
    "window_length" : {DEFAULT : 5,     DESC : "The length of a filter window in samples"},
    "n_clusters" :    {DEFAULT : 10,    DESC : "The number of clusters to use when classifying the data"},
    "offset" :        {DEFAULT : 0,     DESC : "An offset from the start of the data, in samples"},
    "xmin" :          {DEFAULT : None,  DESC : "A minimum value on the x-axis"},
    "xmax" :          {DEFAULT : None,  DESC : "A maximum value on the x-axis"},
    "tmin" :          {DEFAULT : None,  DESC : "A minimum value on the time-axis"},
    "tmax" :          {DEFAULT : None,  DESC : "A maximum value on the time-axis"},
    "commands" :      {DEFAULT : [None],DESC : "TODO"},
    "m" :             {DEFAULT : 1.0,     DESC : "The slope of a linear transform (y = m*x + c)"},
    "c" :             {DEFAULT : 0.0,     DESC : "The intercept of a linear transform (y = m*x + c)"},
    "data_style" :    {DEFAULT : NONE,  DESC : "A string identifier for the data inside the WITSML file"},
    "method" :        {DEFAULT : "lin_fit", DESC : "The method for curve fitting, see scipy.optimize.curve_fit"},
    "mode" :          {DEFAULT : "constant", DESC : "Filter edge handling, see the scipy.ndimage documentation."},
    "moment" :        {DEFAULT : 1,     DESC : "The order of central moment, see scipy.stats.moment"},
    "max_velocity":   {DEFAULT : 1600,  DESC : "The maximum phase velocity"},
    "min_velocity":   {DEFAULT : 1400,  DESC : "The minimum phase velocity"},
    "smooth" :        {DEFAULT : 0, DESC : "The smoothing factor for the filter"},
    "threshold" :     {DEFAULT : 0, DESC : "An upper or lower limit"},
    "value" :         {DEFAULT : 0, DESC : "A single floating point number"},
    "bounds" :        {DEFAULT : ([-5,0],[0,400]), DESC : "The bounds on curve fitting, see scipy.optimize.curve_fit"},
    "initSlopes" :    {DEFAULT : [-0.5], DESC : "An initial slope estimate for the curve fitting, see scipy.optimize.curve_fit"},
    "xsample" :       {DEFAULT : 1,     DESC : "The level of downsampling in the x-directioin"},
    "tsample" :       {DEFAULT : 1,     DESC : "The level of downsampling in the time-direction"},
    "sta":            {DEFAULT : 50,    DESC : "The short-term average window-length in samples"},
    "lta":            {DEFAULT : 200,   DESC : "The long-term average window-length in samples"},
    "xdir" :          {DEFAULT : 5,     DESC : "Wiener filter length in the x-direction"},
    "tdir" :          {DEFAULT : 5,     DESC : "Wiener filter length in the t-direction"},
    "low_freq" :      {DEFAULT : None,  DESC : "A list of low frequency values for band-pass windows in Hz"},
    "high_freq" :     {DEFAULT : None,  DESC : "A list of high frequency values for band-pass windows in Hz"},
    "noisePower" :    {DEFAULT : None,  DESC : "The noise power in the Wiener filter"},      
    "format" :        {DEFAULT : "png", DESC : "The format of the picture output"},
    "clip_level" :    {DEFAULT : 1.0,   DESC : "The number of standard devaitions about the mean for plotting thumbnails"}
    }

    

'''
 DataLoadCommand : the special command, which is number 0 in any workflow
                   requires that some sort of file (in our case a
                   {timestamp}.npy) exists and is loaded so that its data
                   can be processed
'''
class DataLoadCommand(object):
    def __init__(self, data, jsonArgs):
        self._data = data
        self._name = 'load_data'

    def get_name(self):
        return self._name

    # allow for a post condiction that is a
    # list. We use this to return the names of any files
    # that would be written - so that we can test whether or not
    # ateps can be skipped in the RESTART case.
    def postcond(self):
        return []

    def docs(self):
        return {}

    def isGPU(self):
        return False
    
    def execute(self):
        pass

    def result(self):
        return self._data


'''
 All commands that are not the DataLoadCommand must
 inherit from the BasicCommand.

 Override postcond() if you want Restarts to work
 Override docs() if you want automatic self-documentation to work
 Override execute() if you want to do any calculations on the data
'''
# A command that does nothing...
class BasicCommand(object):
    def __init__(self, command, jsonArgs):
        self._previous = command
        self._name = jsonArgs['name']
        self._result = self._previous.result()
        
    def get_name(self):
        return self._name

    # allow for a post condiction that is a
    # list. We use this to return the names of any files
    # that would be written - so that we can test whether or not
    # ateps can be skipped in the RESTART case.
    def postcond(self):
        return []

    def docs(self):
        return {}

    def isGPU(self):
        return False
        
    def execute(self):
        self._result = self._previous.result()
        
        
    def result(self):
        return self._result


'''
 ToGPUCommand : Go to GPU
'''
class ToGPUCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._args = jsonArgs

    def docs(self):
        docs={}
        docs['one_liner']="Put the current result on the GPU and perform subsequent steps there. If no GPU is available, this has no effect."
        return docs

    def isGPU(self):
        return True
    
    def execute(self):
        self._result = extra_numpy.to_gpu(self._previous.result())

'''
 FromGPUCommand : Return from GPU
'''
class FromGPUCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._args = jsonArgs

    def docs(self):
        docs={}
        docs['one_liner']="Put the current result on the CPU and perform subsequent steps there. If no GPU is available, this has no effect."
        return docs

    def isGPU(self):
        return True
    
    def execute(self):
        self._result = extra_numpy.from_gpu(self._previous.result())



'''
  MacroCommand : This implements macros - providing a lightweight system for
                 more complex commands - see extra_numpy.macro_factory for the
                 list of available commands.
'''
class MacroCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._macro = jsonArgs

    def docs(self):
        docs={}
        docs['one_liner']="Create a macro containing sub-commands"
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['command_list'] }
        return docs

    def execute(self):
        traces = self._previous.result()
        # NOTE: the unit-test order matters. 'percentage_on' or 'sum' will result in a single trace, no good for 2D filters...
        command_list = self._macro.get('command_list',[{'name':'abs'}, {'name':'normalize'},
                                                       {'name':'gaussian_filter'},{'name':'gradient'},
                                                       {'name':'threshold'},{'name':'median_filter'},
                                                       {'name':'percentage_on'},{'name':'sum'}])
        for commands in command_list:
            #print(commands,traces.shape)
            traces = extra_numpy.macro_factory(commands, traces)
        self._result = traces

'''
 AbsCommand : wrappers the numpy.abs() function
'''
class AbsCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)

    def docs(self):
        docs={}
        docs['one_liner']="Take the absolute value of the input"
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = extra_numpy.agnostic_abs(self._previous.result())

'''
 AngleCommand : wrappers the numpy.angle() function
'''
class AngleCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)

    def docs(self):
        docs={}
        docs['one_liner']="Take the angle of the complex input"
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = extra_numpy.agnostic_angle(self._previous.result())
'''
 RescaleCommand : rescales the data from 0 to 1
'''
class RescaleCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._prevstack = jsonArgs.get('commands',[None])

    def docs(self):
        docs={}
        docs['one_liner']="Rescale the data from 0 to 1. If an additional data set it suppied, the rescale uses that data for min() and max()"
        return docs

    def isGPU(self):
        return True

    def execute(self):
        Y = None
        if len(self._prevstack)>0:
            if self._prevstack[0] is not None:
                Y = self._prevstack[0].result()
        self._result = extra_numpy.agnostic_rescale(self._previous.result(), external_scaling=Y)
'''
 SoftThresholdCommand : applies a threshold greater or less
'''
class SoftThresholdCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._direction = jsonArgs.get('direction','>')
        self._threshold = jsonArgs.get('threshold',0.0)

    def docs(self):
        docs={}
        docs['one_liner']="Applies a soft threshold, clipping the values at the given threshold."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['direction','threshold'] }
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = extra_numpy.soft_threshold(self._previous.result(),self._threshold,direction=self._direction)

'''
 HardThresholdCommand : applies a threshold greater or less
'''
class HardThresholdCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._direction = jsonArgs.get('direction','>')
        self._threshold = jsonArgs.get('threshold',0.0)
        self._value     = jsonArgs.get('value',0.0)

    def docs(self):
        docs={}
        docs['one_liner']="Applies a hard threshold, all values beyond the threshold are replaced with the supplied value."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['direction','threshold','value'] }
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = extra_numpy.hard_threshold(self._previous.result(),self._threshold, self._value,direction=self._direction)



'''
 UnwrapCommand : unwraps using the numpy.unwrap() function
'''
class UnwrapCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis = jsonArgs.get('axis',-1)

    def docs(self):
        docs={}
        docs['one_liner']="Unwrap angles in the selected axis direction using numpy.unwrap"
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis'] }
        docs['args']['axis']['default']=-1
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = extra_numpy.agnostic_unwrap(self._previous.result(), axis=self._axis)



'''
 ConjCommand : wrappers the numpy.conj() function
'''
class ConjCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)

    def docs(self):
        docs={}
        docs['one_liner']="Take the complex conjugate value of the input"
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = extra_numpy.agnostic_conj(self._previous.result())

'''
 ApproxVlfCommand: weighted average of a trace, gives a reasonable approximation to Very Low Frequency response
'''
class ApproxVlfCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)

    def docs(self):
        docs={}
        docs['one_liner']="A weighted average focussed on the central trace, equivalent to a very low pass filter. The results approximate the very low frequency response in a robust way."
        return docs

    def isGPU(self):
        return False


    def execute(self):
        super().execute()
        self._result = extra_numpy.approx_vlf(self._previous.result())
        

'''
 KerasCommand : wrappers the extra_numpy.kera_model() function
'''
class KerasCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._args = jsonArgs
        self._prevstack = jsonArgs.get('commands',[None])

    def docs(self):
        docs={}
        docs['one_liner']="Load an existing keras model, and either use it for prediction or train-then-predict."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['filename','train'] }
        return docs

    def execute(self):
        train = self._args.get('train',None)
        filename = self._args.get('filename',None)
        Y = None
        if self._prevstack[0] is not None:
            Y = self._prevstack[0].result()
        self._result = extra_numpy.keras_model(self._previous.result(),filename,Y=Y,train=train)

'''
 KMeansCommand : wrappers the extra_numpy.kmeans_clustering() function
'''
class KMeansCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._args = jsonArgs
        self._prevstack = jsonArgs.get('commands',[None])

    def docs(self):
        docs={}
        docs['one_liner']="Perform a KMeans clustering into a fixed number of clusters. Return the cluster number versus depth."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['n_clusters'] }
        return docs

    def execute(self):
        n_clusters = self._args.get('n_clusters',10)
        if len(self._prevstack) < 1:
            self._prevstack = [None]
        if self._prevstack[0] is not None:
            shape_out = self._previous.result().shape
            for a in range(len(self._prevstack)):
                test_shape = self._prevstack[a].result().shape
                if (test_shape[0]*test_shape[1])<(shape_out[0]*shape_out[1]):
                    shape_out=test_shape
            xrows = shape_out[0]*shape_out[1]
            xcols = len(self._prevstack)+1
            data_big = numpy.zeros((xrows,xcols))
            for a in range(xcols-1):
                data_big[:,a]=self._prevstack[a].result().flatten()[:xrows]
            data_big[:,-1]=self._previous.result().flatten()[:xrows]
            result = extra_numpy.kmeans_clustering(data_big,n_clusters=n_clusters)
            self._result = numpy.reshape(result,shape_out)
        else:
            # Single input...
            self._result = extra_numpy.kmeans_clustering(self._previous.result(),n_clusters=n_clusters)



'''
 KurtosisCommand : wrappers the scipy.stats.kurtosis() function
'''
class KurtosisCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis = jsonArgs.get('axis',1)

    def docs(self):
        docs={}
        docs['one_liner']="Take the kurtosis of the input using scipy.stats.kurtosis(). Use k statistics to eliminate bias and omit any NaNs."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis'] }
        return docs

    def execute(self):
        self._result = stats.kurtosis(self._previous.result(),axis=self._axis, bias=False, nan_policy='omit')

'''
 SkewnessCommand : wrappers the scipy.stats.skew() function
'''
class SkewnessCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis = jsonArgs.get('axis',1)

    def docs(self):
        docs={}
        docs['one_liner']="Take the skewness of the input using scipy.stats.skewn(). Eliminate bias and omit any NaNs."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis'] }
        return docs

    def execute(self):
        self._result = stats.skew(self._previous.result(),axis=self._axis, bias=False, nan_policy='omit')

'''
 MomentCommand : wrappers the scipy.stats.moment() function
'''
class MomentCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis = jsonArgs.get('axis',1)
        self._moment = jsonArgs.get('moment',1)

    def docs(self):
        docs={}
        docs['one_liner']="Take the nth moment of the input using scipy.stats.moment(). Omit any NaNs."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis','moment'] }
        return docs

    def execute(self):
        self._result = stats.moment(self._previous.result(),moment=self._moment, axis=self._axis, nan_policy='omit')
'''
 MeanCommand : wrappers the numpy.mean() function
'''
class MeanCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis = jsonArgs.get('axis',1)

    def docs(self):
        docs={}
        docs['one_liner']="Take the mean of the input using numpy.mean()."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis'] }
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = extra_numpy.agnostic_mean(self._previous.result(),axis=self._axis, keepdims=True)
'''
 StdCommand : wrappers the numpy.std() function
'''
class StdDevCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis = jsonArgs.get('axis',1)

    def docs(self):
        docs={}
        docs['one_liner']="Take the standard deviation of the input using numpy.std()."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis'] }
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = numpy.std(self._previous.result(),axis=self._axis, keepdims=True)

'''
 HarmonicMeanCommand : wrappers the scipy.stats.hmean() function
'''
class HarmonicMeanCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis = jsonArgs.get('axis',1)

    def docs(self):
        docs={}
        docs['one_liner']="Take the harmonic mean of the input using scipy.stats.hmean()."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis'] }
        return docs

    def execute(self):
        self._result = stats.hmean(self._previous.result(),axis=self._axis)
'''
 GeometricMeanCommand : wrappers the scipy.stats.gmean() function
'''
class GeometricMeanCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis = jsonArgs.get('axis',1)

    def docs(self):
        docs={}
        docs['one_liner']="Take the geometric mean of the input using scipy.stats.gmean()."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis'] }
        return docs

    def execute(self):
        self._result = stats.gmean(self._previous.result(),axis=self._axis)
'''
 RealCommand : wrappers the numpy.real() function
'''
class RealCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)

    def docs(self):
        docs={}
        docs['one_liner']="Take the real value of the input"
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = extra_numpy.agnostic_real(self._previous.result())

'''
 AnalyticSignalCommand : wrappers the extra_numpy.agnostic_analytic_signal() function
'''
class AnalyticSignalCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._prf = jsonArgs.get('prf',10000)
        self._freq = jsonArgs.get('freq',200.0)
        self._axis = jsonArgs.get('axis',1)
        self._N = jsonArgs.get('window_length',64)
        self._prf = jsonArgs.get('prf',10000.0)
        self._wavn = numpy.abs(jsonArgs['xaxis'][1]-jsonArgs['xaxis'][0])

    def docs(self):
        docs={}
        docs['one_liner']="Estimate the analytic signal using a locally filtered maximum likelihood method"
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis','freq','window_length'] }
        return docs

    def isGPU(self):
        return True

    def execute(self):
        fs = self._prf
        axis = self._axis
        if axis==1:
            fs = self._wavn
        fc = self._freq
        window_length = self._N
        self._result = extra_numpy.agnostic_analytic_signal(self._previous.result(),fc,fs,window_length=window_length, axis=axis)

'''
 HashCommand : creates a 64-bit integer hash using extra_numpy.hash(), reducing the data from 2D to 1D
'''
class HashCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._ioffset = int(jsonArgs.get('offset',0))

    def docs(self):
        docs={}
        docs['one_liner']="Creates a 64-bit integer hash at each depth using extra_numpy.hash(), reducing the data from 2D to 1D"
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['offset'] }
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = extra_numpy.hash(self._previous.result(),ioffset=self._ioffset)
'''
 ButterCommand : wrappers the signal.butter() function and applies it
                 using signal.filtfilt()
'''
class ButterCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._args = jsonArgs
        self._order = jsonArgs.get('order',5)
        self._btype = jsonArgs.get('type','lowpass')
        self._ptype = jsonArgs.get('padtype','even')
        self._prf = jsonArgs.get('prf',10000)
        self._freq = jsonArgs.get('freq',200.0)
        self._axis = jsonArgs.get('axis',1)

    def docs(self):
        docs={}
        docs['one_liner']="Setup a Butterworth filter using scipy.signal.butter() and apply it using scipy.signal.filtfilt()"
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['order','type','padtype','prf','freq','axis'] }
        return docs

    def execute(self):
        Wn = self._freq/(self._prf*0.5)
        if self._axis == 0:
            dx = numpy.abs(self._args['xaxis'][1]-self._args['xaxis'][0])
            Wn = self._freq/(dx*0.5)
        b, a = signal.butter(self._order,Wn,self._btype,output='ba')
        self._result = signal.filtfilt(b,a,self._previous.result(),axis=self._axis, padtype=self._ptype)

'''
 ArgmaxCommand : Used to extract the first index of the maximum value in each
                 row (or column). This is a first order way to extract a
                 profile versus depth from a processed DTS gradient
'''
class ArgmaxCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis = jsonArgs.get('axis',0)

    def docs(self):
        docs={}
        docs['one_liner']="Index of the maximum value in each row or column."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis'] }
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = extra_numpy.agnostic_argmax(self._previous.result(), axis=self._axis)

'''
 GaussianCommand : Applies a 2D Gaussian blurring filter using signal.ndarray.gaussian_filter()
'''
class GaussianCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._xdir = jsonArgs.get('xsigma',1.0)
        self._tdir = jsonArgs.get('tsigma',1.0)
        self._xorder = jsonArgs.get('xorder',5)
        self._torder = jsonArgs.get('torder',5)
        self._mode = jsonArgs.get('mode', 'constant')

    def docs(self):
        docs={}
        docs['one_liner']="Applies a 2D Gaussian blurring filter using signal.ndarray.gaussian_filter()."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['xsigma','tsigma','xorder','torder'] }
        return docs

    def execute(self):
        self._result = ndimage.gaussian_filter(self._previous.result(), [self._xdir, self._tdir], order=[self._xorder, self._torder],mode=self._mode)

'''
 SobelCommand : Applies a Sobel edge detection filter using signal.ndarray.sobel()
'''
class SobelCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis = jsonArgs.get('axis',-1)
        self._mode = jsonArgs.get('mode','constant')

    def docs(self):
        docs={}
        docs['one_liner']="Applies a Sobel edge detection filter using signal.ndarray.sobel()."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis','mode'] }
        return docs

    def execute(self):
        self._result = ndimage.sobel(self._previous.result(),axis=self._axis, mode=self._mode)

'''
 ConvolveCommand : Applies the supplied convolution filter
'''
class ConvolveCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._coeffs = jsonArgs.get('coeffs',[[1,0,-1],[2,0,-2],[1,0,-1]])
        self._mode = jsonArgs.get('mode','constant')

    def docs(self):
        docs={}
        docs['one_liner']="Convolves the supplied filter with the data."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['coeffs','mode'] }
        docs['args']['coeffs']['default'] = [[1,0,-1],[2,0,-2],[1,0,-1]] 
        return docs

    def execute(self):
        self._result = ndimage.convolve(self._previous.result(),self._coeffs, mode=self._mode)

'''
 CorrelateCommand : Applies the supplied correlation filter
'''
class CorrelateCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._coeffs = jsonArgs.get('coeffs',[[0,0,1],[0,1,0],[1,0,0],[0,1,0],[0,0,1]])
        self._mode = jsonArgs.get('mode','constant')

    def docs(self):
        docs={}
        docs['one_liner']="Correlates the supplied filter with the data."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['coeffs','mode'] }
        docs['args']['coeffs']['default'] = [[0,0,1],[0,1,0],[1,0,0],[0,1,0],[0,0,1]]
        return docs

    def execute(self):
        self._result = ndimage.correlate(self._previous.result(),self._coeffs, mode=self._mode)



'''
  MedianFilterCommand : applies a 2D square median filter using ndimage.median_filter()
'''
class MedianFilterCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._distance = jsonArgs.get('distance',3)

    def docs(self):
        docs={}
        docs['one_liner']="Applies a 2D square median filter using ndimage.median_filter()."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['distance'] }
        return docs

    def execute(self):
        self._result = ndimage.median_filter(self._previous.result(), self._distance)


'''
 SumCommand : Sums data along the specified axis, reducing from 2D to 1D using numpy.sum()
'''
class SumCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis = jsonArgs.get('axis',1)

    def docs(self):
        docs={}
        docs['one_liner']="Sums the data along the specified axis, reducing from 2D to 1D using numpy.sum()."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis'] }
        docs['args']['axis']['default']=1
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = extra_numpy.agnostic_sum(self._previous.result(), axis=self._axis, keepdims=True)

'''
 RollCommand : rolls the data along the requested axis using extra_numpy.agnostic_roll
'''
class RollCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis  = jsonArgs.get('axis',1)
        self._shift = jsonArgs.get('window_length',1)

    def docs(self):
        docs={}
        docs['one_liner']="Rolls the data along the specified axis, using numpy.roll()."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis','window_length'] }
        docs['args']['axis']['default']=1
        docs['args']['window_length']['default']=1
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = extra_numpy.agnostic_roll(self._previous.result(), self._shift, axis=self._axis)





'''
 RunningMeanCommand : A 1D running mean averaging filter, using the extra_numpy.running_mean()
'''
class RunningMeanCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        # previously was 'length'
        self._N = jsonArgs.get('window_length',5)
        self._axis = jsonArgs.get('axis',0)

    def docs(self):
        docs={}
        docs['one_liner']="A 1D running mean averaging filter, using the extra_numpy.running_mean()."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['window_length','axis'] }
        docs['args']['axis']['default']=0
        return docs

    def execute(self):
        self._result = extra_numpy.running_mean(self._previous.result(), self._N, axis=self._axis)


'''
 InterpCommand : interpolation for downsampling using extra_numpy.interp()
'''
class InterpCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._xp = jsonArgs.get('xaxis', None)
        self._fp = jsonArgs.get('fp', None)
        self._axis = jsonArgs.get('axis',0)

    def docs(self):
        docs={}
        docs['one_liner']="Interplation for downsampling."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['xaxis','fp','axis'] }
        docs['args']['axis']['default']=0
        return docs

    def execute(self):
        self._result = extra_numpy.interp(self._previous.result(), self._xp, self._fp, axis=self._axis)

'''
 CurveFitCommand : Flexible curve fitting using scipy.optimize.curvefit()
'''
class CurveFitCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._func = jsonArgs.get('method','lin_fit')
        self._bound = jsonArgs.get('bounds',([-5,0],[0,400]))
        self._initslope = jsonArgs.get('initSlopes',[-0.5])
        self._xaxis = jsonArgs.get('xaxis',None)

    def docs(self):
        docs={}
        docs['one_liner']="Curve fitting using scipy.optimize.curvefit()."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['method','bounds','initSlopes','xaxis'] }
        return docs

    def execute(self):
        fit_func, coeffs0 = extra_numpy.select_fit(self._func, self._previous.result(),self._initslope)
        coeffs,pcov = curve_fit(fit_func,self._previous.result(),self._xaxis,p0=coeffs0,
                                      bounds=self._bounds)
        self._result = fit_func(self._previous.result(), coeffs)
    
'''
 ClipCommand : clip the current matrix so that further operations operate on a smaller window
'''
class ClipCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._dirname = jsonArgs.get('directory_out','NONE')
        self._xaxis = jsonArgs.get('xaxis',None)
        self._taxis = jsonArgs.get('taxis',None)            
        self._xmin = jsonArgs.get('xmin',numpy.min(self._xaxis))
        self._xmax = jsonArgs.get('xmax',numpy.max(self._xaxis))
        self._tmin = jsonArgs.get('tmin',numpy.min(self._taxis))
        self._tmax = jsonArgs.get('tmax',numpy.max(self._taxis))
        self._double_ended = jsonArgs.get('double_ended',-1)
        # for unit tests and documentation...
        if self._tmin is None:
            self._tmin=0
        self._fname = str(numpy.int(self._tmin))

    def docs(self):
        docs={}
        docs['one_liner']="Clip the data so that all subsequent operations operate on a small window. If directory_out is specified the new axes will also be created in the storage directory"
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['xaxis','taxis','xmin','xmax','tmin','tmax','directory_out','double_ended'] }
        return docs

    def isGPU(self):
        return True

    def postcond(self):
        return [directory_services.path_join(self._dirname,self._fname+'.npy')]

    def execute(self):
        data = self._previous.result()
        xend = int(self._xaxis.shape[0]*0.5)
        print(xend)
        if self._double_ended==0:
            self._xaxis = self._xaxis[:xend]
            data = data[:xend,:]
        if self._double_ended==1:
            self._xaxis = self._xaxis[xend:]
            data = data[xend:,:]
            self._xaxis = self._xaxis[::-1]
            data = data[::-1,:]
        ixmin = (numpy.abs(self._xaxis-self._xmin)).argmin()
        ixmax = (numpy.abs(self._xaxis-self._xmax)).argmin()
        itmin = 0
        itmax = 0
        if self._taxis is not None:
            itmin = (numpy.abs(self._taxis-self._tmin)).argmin()
            itmax = (numpy.abs(self._taxis-self._tmax)).argmin()
        else:
            itmin=0
            itmax=data.shape[1]
        print('XMIN',self._xmin,self._xaxis[ixmin], ixmin)
        print('XMAX',self._xmax,self._xaxis[ixmax], ixmax)
        self._result = (data)[ixmin:ixmax,itmin:itmax]
        dirname = self._dirname
        if not dirname=='NONE':
            io_helpers.numpy_out(dirname,self._fname,self._result)
            io_helpers.numpy_out(dirname,'measured_depth',self._xaxis[ixmin:ixmax])
            io_helpers.numpy_out(dirname,'time',self._taxis[itmin:itmax])
'''
 DiffCommand : numerical differencing with a window_length offset
'''
class DiffCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis = jsonArgs.get('axis',-1)
        # previously was 'n'
        self._n = jsonArgs.get('window_length',1)

    def docs(self):
        docs={}
        docs['one_liner']="Simple differencing using the window_length as the offset."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['window_length','axis'] }
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = extra_numpy.agnostic_diff(self._previous.result(),n=self._n, axis=self._axis)
        # boundary condition
        if self._axis==1:
            self._result[:,-1]=self._result[:,-2]
        if self._axis==0:
            self._result[-1,:]=self._result[-2,:]

'''
 GradientCommand : numerical central differencing
'''
class GradientCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._xaxis = jsonArgs['xaxis']
        self._taxis = jsonArgs.get('taxis',None)            
        self._edge_order = jsonArgs.get('edge_order',1)
        self._axis = jsonArgs.get('axis',0)

    def docs(self):
        docs={}
        docs['one_liner']="Numercal gradient of the data via central differencing."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['edge_order','axis'] }
        docs['args']['axis']['default']=0
        return docs

    def isGPU(self):
        return False

    def execute(self):
        if self._axis is not None:
            if self._axis==1:
                self._result = numpy.gradient(self._previous.result(),self._taxis,edge_order=self._edge_order, axis=1)
            else:
                self._result = numpy.gradient(self._previous.result(),self._xaxis,edge_order=self._edge_order, axis=0)
        else:
            self._result = numpy.gradient(self._previous.result(),[self._xaxis,self._taxis],edge_order=self._edge_order)

                

'''
 DownsampleCommand : index-based downsampling. Note there is no filtering, so for anti-alias
                     downsampling apply an appropriate low-pass filter before calling this.
'''
class DownsampleCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._xdir = jsonArgs.get('xsample',1)
        self._tdir = jsonArgs.get('tsample',1)

    def docs(self):
        docs={}
        docs['one_liner']="Downsampling to reduce the data size."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['xsample','tsample'] }
        return docs

    def isGPU(self):
        return True


    def execute(self):
        self._result = (self._previous.result())[::self._xdir,::self._tdir]

'''
 Peak2PeakCommand : The maximum peak-to-peak difference with the maximum and minimum separated
                    by less than the defined window_length. This reduces the data from 2D (x,t)
                    to a trace (x) using extra_numpy.peak_to_peak()
'''
class Peak2PeakCommand(BasicCommand):
    def __init__(self,command,jsonArgs):
        super().__init__(command, jsonArgs)
        self._window_length = jsonArgs.get('window_length',200)

    def docs(self):
        docs={}
        docs['one_liner']="The maximum peak-to-peak difference with the maximum and minimum separated by less than the defined window_length. This reduces the data from 2D (x,t) to a trace (x)."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['window_length'] }
        return docs


    def execute(self):
        self._result = extra_numpy.peak_to_peak(self._previous.result(), self._window_length)

'''
 CountPeaksCommand : Counts peaks using extra_numpy.count_peaks(), which uses scipy.signal.find_peaks_cwt()
'''
class CountPeaksCommand(BasicCommand):
    def __init__(self,command,jsonArgs):
        super().__init__(command, jsonArgs)
        self._sta = jsonArgs.get('sta',50)
        self._lta = jsonArgs.get('lta',200)

    def docs(self):
        docs={}
        docs['one_liner']="Couting peaks in a signal using scipy.signal.find_peaks_cwt()."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['sta','lta'] }
        return docs


    def execute(self):
        self._result = extra_numpy.count_peaks(self._previous.result(), self._sta, self._lta)


'''
 MultiplyCommand : elementwise multiply
'''
class MultiplyCommand(BasicCommand):
    def __init__(self,command,jsonArgs):
        super().__init__(command, jsonArgs)
        self._prevstack = jsonArgs.get('commands',[None])

    def docs(self):
        docs={}
        docs['one_liner']="Elementwise multiply, the output data-type will be the same as that of the data entering in the in_uid. This data is multiplied by data provided in the gather_uids"
        #docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['window_length','axis'] }
        return docs

    def isGPU(self):
        return True

    def execute(self):
        if self._prevstack[0]!=None:
            self._result = self._previous.result()*((self._prevstack[0]).result())
'''
 AddCommand : elementwise sum
'''
class AddCommand(BasicCommand):
    def __init__(self,command,jsonArgs):
        super().__init__(command, jsonArgs)
        self._prevstack = jsonArgs.get('commands',[None])

    def docs(self):
        docs={}
        docs['one_liner']="Elementwise sum, the output data-type will be the same as that of the data entering in the in_uid."
        #docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['window_length','axis'] }
        return docs

    def isGPU(self):
        return True

    def execute(self):
        if self._prevstack[0]!=None:
            self._result = self._previous.result()+((self._prevstack[0]).result())
'''
 LinearScalarCommand : y = mx+c
'''
class LinearTransformCommand(BasicCommand):
    def __init__(self,command,jsonArgs):
        super().__init__(command, jsonArgs)
        self._m = jsonArgs.get('m', 1.0)
        self._c = jsonArgs.get('c', 0.0)

    def docs(self):
        docs={}
        docs['one_liner']="Provide two scalars m and c, linearly transform the data y = m*data + c"
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['m','c'] }
        return docs

    def isGPU(self):
        return True

    def execute(self):
        m = self._m
        c = self._c
        self._result = m*self._previous.result()+c


'''
 GatherCommand : Gather results together
'''
class GatherCommand(BasicCommand):
    def __init__(self,command,jsonArgs):
        super().__init__(command, jsonArgs)
        self._prevstack = jsonArgs.get('commands',[None])

    def docs(self):
        docs={}
        docs['one_liner']="Gathers the data with all the data provided in the gather_uids to make one big matrix"
        return docs

    def isGPU(self):
        return True

    def execute(self):
        if self._prevstack[0]!=None:
            datalist = []
            for dataset in self._prevstack:
                datalist.append(dataset.result())
            self._result = extra_numpy.gather(self._previous.result(),datalist)

'''
 StaLtaCommand : short-term average to long-term-average onset-picker transform
'''
class StaLtaCommand(BasicCommand):
    def __init__(self,command,jsonArgs):
        super().__init__(command, jsonArgs)
        self._sta = jsonArgs.get('sta',50)
        self._lta = jsonArgs.get('lta',200)

    def docs(self):
        docs={}
        docs['one_liner']="Short-term average (STA) divided by long-term average (LTA). This transform highlights onset and so often forms part of an automated pick or edge-detection."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['sta','lta'] }
        return docs


    def execute(self):
        self._result = extra_numpy.sta_lta(self._previous.result(), self._sta, self._lta)

'''
 WienerCommand : Apply a 2D Wiener filter using scipy.signal.wiener()
'''
class WienerCommand(BasicCommand):
    def __init__(self,command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._xdir = jsonArgs.get('xdir',5)
        self._tdir = jsonArgs.get('tdir',5)
        self._noisePower = jsonArgs.get('noisePower',None)

    def docs(self):
        docs={}
        docs['one_liner']="2D Wiener filter. See scipy.signal.wiener."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['xdir','tdir','noisePower'] }
        return docs


    def execute(self):
        print('wiener ['+str(self._xdir) + ', ' + str(self._tdir) + ']')
        self._result = signal.wiener(self._previous.result(), [self._xdir, self._tdir], self._noisePower)


'''
 WriteNPYCommand : writes out the current datablock to NPY format
'''
class WriteNPYCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._dirname = jsonArgs.get('directory_out','NONE')
        self._fname = jsonArgs.get('date_dir','NONE')
        self._xaxis = jsonArgs['xaxis']
        self._taxis = jsonArgs.get('taxis',None)            


    def docs(self):
        docs={}
        docs['one_liner']="Write the current state of the processed data to the npy format."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['directory_out'] }
        return docs

    def postcond(self):
        return [directory_services.path_join(self._dirname,self._fname+'.npy')]

    def isGPU(self):
        return True

    def execute(self):
        super().execute()
        dirname = self._dirname
        if not dirname=='NONE':
            extra_numpy.agnostic_save(dirname,self._fname,self._previous.result())
            extra_numpy.agnostic_save(dirname,'measured_depth',self._xaxis)
            if self._taxis is not None:
                extra_numpy.agnostic_save(dirname,'time',self._taxis)

'''
 WriteNPYCommand : writes out the current datablock to NPY format
'''
class ReadNPYCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._dirname = jsonArgs.get('directory_out','NONE')
        self._fname = jsonArgs.get('filename','NONE')

    def docs(self):
        docs={}
        docs['one_liner']="Load a npy format file."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['directory_out','filename'] }
        return docs

    def isGPU(self):
        return True

    def execute(self):
        super().execute()
        dirname = self._dirname
        if not dirname=='NONE':
            # we can use the previous result to tell us whether we are on GPU or CPU
            self.result = extra_numpy.agnostic_load(dirname,self._fname,self._previous.result())


'''
 WriteWITSMLCommand : writes the results to the WITSML/FBE format.
'''
class WriteWITSMLCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._outdir = jsonArgs.get('directory_out','NONE')
        self._xaxis = jsonArgs.get('xaxis',None)
        self._prf = jsonArgs.get('prf',10000)
        self._prevstack = jsonArgs.get('commands',[None])
        self._data_style = jsonArgs.get('data_style','NONE')
        self._datedir = jsonArgs.get('date_dir','NONE')
        self._datestring = jsonArgs.get('datestring','NONE')
        self._lowf = jsonArgs.get('low_freq',[0])
        self._highf = jsonArgs.get('high_freq',[1])
        self._labellist = jsonArgs.get('labels',[])
        #print(jsonArgs)

    def postcond(self):
        return [directory_services.path_join(self._outdir,self._datedir+'.fbe')]


    def docs(self):
        docs={}
        docs['one_liner']="Write out to the WITSML/FBE format, suitable for loading into viewers such as Techlog or Petrel."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['directory_out','xaxis','data_style','low_freq','high_freq','labels'] }
        return docs


    def execute(self):
        if not self._outdir=='NONE':
            # Resolve band00 here - because if you do it in __init__ you
            # may not have executed the Command that generates the result
            # and so you will get the input a
            self._band00 = None
            # Here we extract the result...because we never pass commands outside
            # the command_set..only numpy or similar in the calculations themselves
            if self._prevstack[0]!=None:
                self._band00= (self._prevstack[0]).result()
            io_helpers.write2witsml(self._outdir,self._datedir,self._datestring,self._xaxis, self._band00, self._previous.result(),
                                    self._lowf, self._highf, self._prf, data_style=self._data_style, label_list=self._labellist)

'''
 RMSfromFFTCommand : the calculation of RMS noise within a frequency band, using the reduced memory functions
'''
class RMSfromFFTCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)
        f_rescale = jsonArgs.get('f_rescale',1.0)
        self._i1 = int(jsonArgs.get('low_freq',0)*f_rescale)
        highf = jsonArgs.get('high_freq',-1)
        if highf>0:
            highf = highf*f_rescale
        if (int(highf)==self._i1):
            highf=-1
        self._i2 = int(highf)

    def docs(self):
        docs={}
        docs['one_liner']="Calculate the RMS energy between two frequencies."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['low_freq','high_freq'] }
        return docs

    def isGPU(self):
        return True
    

    def execute(self):
        i1 = self._i1
        i2 = self._i2
        if i2<1:
            i2 = int(self._previous.result().shape[1]/2)
        self._result = extra_numpy.agnostic_zeros(self._previous.result(),(self._previous.result().shape[0],1),dtype=numpy.double)
        self._result = extra_numpy.reduced_mem((self._previous.result()[:,i1:i2]), self.result(), extra_numpy.rms_from_fft, {})


'''
 FFTCommand : applies and FFT along the requested axis using the reduced memory functions
'''
class FFTCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis = jsonArgs.get('axis',1)

    def docs(self):
        docs={}
        docs['one_liner']="Compute the Fast Fourier Transform (FFT) of the data along the requested axis."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis'] }
        return docs

    def isGPU(self):
        return True
    

    def execute(self):
        if (self._axis==1):
            self._result = extra_numpy.agnostic_zeros(self._previous.result(),(self._previous.result().shape), dtype=numpy.complex)
            self._result = extra_numpy.reduced_mem(self._previous.result(), self.result(), extra_numpy.agnostic_fft, {'axis':self._axis})
        else:
            # currently the low memory version is only developed for the time orientation....
            self._result = extra_numpy.agnostic_fft(self._previous.result(),axis=self._axis)

'''
 IFFTCommand : the inverse FFT along the requested axis
'''
class IFFTCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._axis = jsonArgs.get('axis',1)

    def docs(self):
        docs={}
        docs['one_liner']="Compute the Inverse Fast Fourier Transform (IFFT) of the data along the requested axis."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis'] }
        return docs

    def isGPU(self):
        return True

    def execute(self):
        if (self._axis==1):
            self._result = extra_numpy.agnostic_zeros(self._previous.result(),(self._previous.result().shape), dtype=numpy.complex)
            self._result = extra_numpy.reduced_mem(self._previous.result(), self.result(), extra_numpy.agnostic_ifft, {'axis':self._axis})
        else:
            # currently the low memory version is only developed for the time orientation....
            self._result = extra_numpy.agnostic_ifft(self._previous.result(),axis=self._axis)

'''
 DesripeCommand : Remove vertical stripes from the data using extra_numpy.destripe()
'''
class DestripeCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)

    def docs(self):
        docs={}
        docs['one_liner']="Remove vertical stripes from the data using extra_numpy.destripe()."
        return docs

    def isGPU(self):
        return True
    
    def execute(self):
        self._result = extra_numpy.destripe(self._previous.result())

'''
 UpCommand : Calculate the up-going waves using extra_numpy.up_wave()
'''
class UpCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)

    def docs(self):
        docs={}
        docs['one_liner']="Calculate the up-going waves using extra_numpy.up_wave(). Note that the data should be 2D FFTd before this command and are returned as complex values."
        return docs

    def isGPU(self):
        return True    

    def execute(self):
        self._result = extra_numpy.up_wave(self._previous.result())

'''
 DownCommand : Calculate the down-going waves using extra_numpy.down_wave()
'''
class DownCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)

    def docs(self):
        docs={}
        docs['one_liner']="Calculate the down-going waves using extra_numpy.down_wave(). Note that the data should be 2D FFTd before this command and are returned as complex values."
        return docs

    def isGPU(self):
        return True

    def execute(self):
        self._result = extra_numpy.down_wave(self._previous.result())

'''
 VelocityMapCommand : Calculate the phase-velocity at each pixel in a 2D FFT space
'''
class VelocityMapCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._args = jsonArgs

    def docs(self):
        docs={}
        docs['one_liner']="Calculate the phase velocity at each pixel in a 2D FFT space."
        return docs


    def execute(self):
        localshape = self._previous.result().shape
        nx = localshape[0]
        nt = localshape[1]
        dx = numpy.abs(self._args['xaxis'][1]-self._args['xaxis'][0])
        dt = 1.0/self._args['prf']
        self._result = extra_numpy.vel_map(nx,nt,dx,dt)

'''
 VelocityMaskCommand : Construct a phase-velocity filter in 2D space. The input should be from the velocity_map command
'''
class VelocityMaskCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)
        self._args = jsonArgs

    def docs(self):
        docs={}
        docs['one_liner']="Construct a phase-velocity filter in 2D space. The input should be from the velocity_map command."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['min_velocity','max_velocity'] }
        return docs

    def execute(self):
        velmap = self._previous.result()
        minv = self._args.get('min_velocity',1400.0)
        maxv = self._args.get('max_velocity', 1600.0)
        smoothval = self._args.get('smooth',0)
        self._result = extra_numpy.vel_mask(velmap,minv,maxv,smoothval)

'''
 MultipleCalcsCommand : allows multiple calculations to be made using the reduced_mem form.  
'''
class MultipleCalcsCommand(BasicCommand):
    def __init__(self, command, jsonArgs):
        super().__init__(command, jsonArgs)
        f_rescale = jsonArgs.get('f_rescale',1.0)
        flist = jsonArgs.get('low_freq',[0])
        self._lowf = []
        for freq in flist:
            self._lowf.append(int(freq*f_rescale))
        flist = jsonArgs.get('high_freq',[0])
        self._highf = []
        icount =0
        for freq in flist:
            highf=freq
            if highf>0:
                highf=highf*f_rescale
            if int(highf)==self._lowf[icount]:
                highf=-1
                print("RMS band too narrow - setting to Nyquist")
            self._highf.append(int(highf))
            icount=icount+1
        self._funcName = jsonArgs.get('func','NONE')

    def docs(self):
        docs={}
        docs['one_liner']="Perform multiple calculations using the extra_numpy.reduced_mem() system."
        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['low_freq','high_freq','func'] }
        return docs

    def isGPU(self):
        return True
        
    def execute(self):
        if not self._funcName=='NONE':
            nyq = int(self._previous.result().shape[1]/2)
            for a in range(len(self._highf)):                
                if (self._highf[a]<0):
                    self._highf[a] = nyq
            self._result = extra_numpy.multiple_calcs(self._previous.result(),self._lowf, self._highf, extra_numpy.available_funcs(self._funcName))

'''
 The KnownCommands. If a command does not appear here it can't be accessed from distpy
'''
def KnownCommands(knownList):
    knownList['NONE']           = BasicCommand
    knownList['data']           = DataLoadCommand
    knownList['abs']            = AbsCommand
    knownList['angle']          = AngleCommand
    knownList['add']            = AddCommand
    knownList['analytic_signal']= AnalyticSignalCommand
    knownList['argmax']         = ArgmaxCommand
    knownList['approx_vlf']     = ApproxVlfCommand
    knownList['butter']         = ButterCommand
    knownList['clip']           = ClipCommand
    knownList['conj']           = ConjCommand
    knownList['convolve']       = ConvolveCommand
    knownList['correlate']      = CorrelateCommand
    knownList['count_peaks']    = CountPeaksCommand
    knownList['data_load']      = ReadNPYCommand
    knownList['destripe']       = DestripeCommand
    knownList['diff']           = DiffCommand
    knownList['downsample']     = DownsampleCommand
    knownList['down_wave']      = DownCommand
    knownList['fft']            = FFTCommand
    knownList['from_gpu']       = FromGPUCommand
    knownList['gather']         = GatherCommand
    knownList['gaussian']       = GaussianCommand
    knownList['geometric_mean'] = GeometricMeanCommand
    knownList['gradient']       = GradientCommand
    knownList['harmonic_mean']  = HarmonicMeanCommand
    knownList['hard_threshold'] = HardThresholdCommand
    knownList['ifft']           = IFFTCommand
    knownList['keras']          = KerasCommand
    knownList['kmeans']         = KMeansCommand
    knownList['kurtosis']       = KurtosisCommand
    knownList['lin_transform']  = LinearTransformCommand
    knownList['macro']          = MacroCommand
    knownList['median_filter']  = MedianFilterCommand
    knownList['mean']           = MeanCommand
    knownList['multiply']       = MultiplyCommand
    knownList['multiple_calcs'] = MultipleCalcsCommand
    knownList['peak_to_peak']   = Peak2PeakCommand
    knownList['rms_from_fft']   = RMSfromFFTCommand
    knownList['real']           = RealCommand
    knownList['rescale']        = RescaleCommand
    knownList['roll']           = RollCommand
    knownList['running_mean']   = RunningMeanCommand
    knownList['sobel']          = SobelCommand
    knownList['soft_threshold'] = SoftThresholdCommand
    knownList['sum']            = SumCommand
    knownList['skewness']       = SkewnessCommand
    knownList['sta_lta']        = StaLtaCommand
    knownList['std_dev']        = StdDevCommand
    knownList['to_gpu']         = ToGPUCommand
    knownList['unwrap']         = UnwrapCommand
    knownList['up_wave']        = UpCommand
    knownList['velocity_map']   = VelocityMapCommand
    knownList['velocity_mask']  = VelocityMaskCommand
    knownList['wiener']         = WienerCommand
    knownList['write_npy']      = WriteNPYCommand
    knownList['write_witsml']   = WriteWITSMLCommand
    return knownList    

'''
 tesxt_for_users : A header and introduction for autogenerated documentation.
'''
def text_for_users():
    text_for_users = {}
    text_for_users['heading']="Public Command Set"
    text_for_users['text'] = [("A set of commands that can be collected together to form a tree-structure ",
                              "for signal processing. The general form of a command is to have a unique identifier ",
                              " \"uid\" and to inherit from a parent via that parent's uid \"uid_in\".\n",
                              "\n",
                              "Commands may have additional attributes as outlined below:\n")]
    
    



import os
from setuptools import setup, find_packages
from distpy import __version__, __license__
'''
 distpy : module for supporting signal processing workflows for DAS/DTS
          across multiple platforms.

VERSIONING
===========
Semantic versioning follows the pattern of ‹major›.‹minor›.‹patch›.

Rules:
Major : only for backwards incompatible changes
Minor : if not Major and not Patch
Patch : if and only if there is a fix that does not change functionality

The following story shows how the version name might evolve:
0.0.0 - first version
...Someone finds a bug, it gets fixed: 0.0.1
...We change the name of the controllers so that the old names won't work anymore 1.0.1
...We add some features to the command set so that you have more processing options 1.1.1


BUILD
=====
python -m setup sdist

INSTALL
=======
python -m setup install
      packages=['calc', 'controllers', 'ingesters', 'io_help','workers'],
      package_dir={                                                                                                                                                                                                
              'calc'       : 'distpy\calc',                                                                                                                                                                                       
              'controllers': 'distpy\controllers',                                                                                                                                                                              
              'ingesters'  : 'distpy\ingesters',                                                                                                                                                                            
              'io_help'    : 'distpy\io_help',
              'workers'    : 'distpy\workers'},             
      package_dir={'': 'distpy',
                   'distpy.io_help' : 'distpy\io_help'},
      packages=find_namespace_packages(where='distpy'),
          package_dir = {
              'distpy' : 'distpy',
              'distpy.calc' : 'distpy\calc',
              'dispty.controllers': 'distpy\controllers',
              'distpy.ingesters'  : 'distpy\ingesters',
              'distpy.io_help'    : 'distpy\io_help',
              'distpy.workers'    : 'distpy\workers'},

'''
import builtins

# Get a list of all files in the JS directory to include in our module
builtins.__DISTPY_SETUP__ = True
def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths
py_files = package_files('distpy')

with open("README.md", "r") as fh:
    long_description = fh.read()

def setup_package():
    setup(name='distpy',
          version=__version__,
          description='distpy : Processing for distributed fibre-optic sensor data',
          long_description=long_description,
          long_description_content_type="text/markdown",
          author="Michael J. Williams",
          author_email="miw@slb.com",
          packages = find_packages(),
          package_data={"distpy":py_files},
          license=__license__,
          keywords = ["DAS", "DTS", "hDVS", "DVS"],
          url="https://github.com/Schlumberger/distpy",
          classifiers=[
              'Development Status :: 2 - Pre-Alpha',
              'Environment :: Other Environment',
              'Intended Audience :: Developers',
              'Intended Audience :: Science/Research',
              'License :: OSI Approved :: MIT License',
              'Natural Language :: English',
              'Programming Language :: Python',
              'Programming Language :: Python :: 3.7',
              'Topic :: Scientific/Engineering',
              'Topic :: Scientific/Engineering :: Physics',
              'Topic :: Software Development :: Libraries',
              'Topic :: Utilities'
          ],
          install_requires=[
              'numpy',
              'scipy',
              'numba',
              'matplotlib',
              'h5py',
              'pandas',
              'scikit-learn'
              ],
          zip_safe = False
           )

if __name__ == '__main__':
    setup_package()
    del builtins.__DISTPY_SETUP__



# (C) 2020, Schlumberger. Refer to LICENSE
#
# unit_handler.py
#
# NOTE: units should only be used by plots, we allow an assumption of
#       metric everywhere for all calculations and processing.


from numba import vectorize, float64


#UNITS FOR DISPLAY
@vectorize([float64(float64)],)
def M_TO_FT(val):
    return 3.28084*val
@vectorize([float64(float64)])
def FT_TO_M(val):
    return val/3.28084

# NOTE: approximate temperature conversion
#        don't go C_TO_F(F_TO_C(C_TO_F())) etc. as there will be drift
@vectorize([float64(float64)])
def C_TO_F(val):
    return (1.4*val)+32.0
@vectorize([float64(float64)])
def F_TO_C(val):
    return (5.0/9.0)*(val-32.0)



from __future__ import division, absolute_import, print_function

__all__ = ['io_help','calc','controllers','ingesters','workers']

__copyright__ = '(C) 2020, Schlumberger. Refer to LICENSE'
__license__ = '../LICENSE'
__status__ = 'Pre-alpha'
# version tracking
# 0.0.0 Baseline version containing the framework for the system.
# 0.1.0 Added the clip command, with a unit-test and documented in the Wiki.
# 0.2.0 Added the diff command, with a unit-test and documented in the Wiki.
# 0.3.0 Added the averageFBE command, with a unit-test and documented in the Wiki.
# 0.3.1 Minor bug fixes to make sure that the RAW-to-seismic integration tests runs
# 0.3.2 Minor bug-fixes to make sure that strainrate to WITSML processing works
# 0.3.3 Bug-fix to ensure directory creation when ingesting SEGY
# 1.0.0 change to the __init__.py format for Azure compatibility
# 1.1.0 post-processing and plotting capabilities added
# 1.2.0 source and reflection profiling
# 1.3.0 basic event detector using peak_to_peak(sta_lta(abs()))
# 1.3.1 attempt to improve accuracy of ibm2ieee conversion
# 1.4.0 velocity maps and masks to enable Velocity Band Energy workflows
# 1.4.1 faster calculation for backscatter-to-strainrate part. Performance tests.
# 1.4.2 - fix so that axes on the thumbnail plot are physical axes
#       - allowed "format" to specified in thumbnail command for jpg option
#       - introduced "real" command to allow the results of velocity masks to be consumed
#       - bug-fix to the velocity mask as the intepretation of the input keys was wrong
#       - sped up the calculation for the sta/lta and the peak-to-peak
#       - introduced wavelet based peak-counting
# 1.5.0 DTS support, in particular the steam-injection workflow starting from
#       CSV data
# 1.5.1 - Support for a scaling inside the thumbnail - this is the number of std devs to bracket the image's colorscale
# 1.5.2 - Bug-fix in the calculation of the diff-filter for poly-pulse processing, to avoid a possible divide-by-zero error.
# 1.6.0 Virtual sources, sweetness attribute and the gauge-length stacking
# 1.6.1 Clipping for data that has not got Time_axis defined
# 1.7.0 destripe filter and auto-docoumentation
# 1.8.0 read h5, read CSV...
# 1.9.0 hash command added for perceptual hash workflows
# 1.10.0 Support for proprietary extensions
# 1.10.1 Extra protection against divide by zero in sgy.py
# 1.11.0 Enabled the basic statistical summaries
# 1.12.0 Enabled the GPU options
# 1.13.0 Added an additional h5 reader option
# 1.14.0 Added the keras command
# 1.14.1 Corrected the auto-generated documentation on the keras command
# 1.15.0 Added the kmeans command
# 1.15.1 Corrected the auto-generated documentation on the kmeans command
# 1.15.2 Corrected setup.py so that all necessary libraries are installed
# 1.16.0 Added the approximate very low frequency (VLF) calculation
# 1.16.1 Corrected the automatic documentation for the analytic signal command
# 1.16.2 corrected the time.npy output file from the CSV ingester.
# 1.17.0 linear transform and addition included to make custom filters easier to design
# 1.18.0 Generalized the kmeans clustering
# 1.19.0 CSV ingestion to WITSML
# 1.20.0 CSV ingestion from DTS-style (depth as rows, time as columns)
# 1.21.0 Improvement in data-handling of difference cases so that they can be plotted
# 1.22.0 Improvement in self documenting features
# 1.23.0 GPU write_numpy to make hybrid cloud easier as a deployment
# 1.23.1 Fix a bug in write_numpy
# 1.24.0 Allowed an external data-set to be used for min() and max() in the rescale command
# 1.24.1 Improved behaviour when an unknown command is tried
__version__ = '1.24.1'


